import winreg
import ctypes
import sys
from enum import Enum



# ------------------------------------------------------------------

def is_admin():
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        return False

def run_admin():
    if not is_admin():
        ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, sys.argv[0], None, 1)
        sys.exit()

# ------------------------------------------------------------------

def create_key(path, hive = winreg.HKEY_CLASSES_ROOT):
    winreg.CreateKey(hive, path)

def set_key_value(key_path, subkey_name, value, hive=winreg.HKEY_CLASSES_ROOT):
    '''
    Changes the value of a subkey.

    Args: Path to key, name of subkey, and value
    '''

    registry_key = winreg.OpenKey(hive, key_path, 0,
                                  winreg.KEY_WRITE)
    winreg.SetValueEx(registry_key, subkey_name, 0, winreg.REG_SZ, value)
    winreg.CloseKey(registry_key)


def list_keys(path, hive = winreg.HKEY_CLASSES_ROOT):

    open_key = winreg.OpenKey(hive, path)
    key_amt = winreg.QueryInfoKey(open_key)[0]
    keys = []

    for count in range(key_amt):
        subkey = winreg.EnumKey(open_key, count)
        keys.append(subkey)

    return keys

def delete_key(path, hive = winreg.HKEY_CLASSES_ROOT):
    open_key = winreg.OpenKey(hive, path)
    subkeys = list_keys(path)

    if len(subkeys) > 0:
        for key in subkeys:
            delete_key(path + '\\' + key)
    winreg.DeleteKey(open_key, "")



# ------------------------------------------------------------------



# run_admin()
# print(list_keys('Directory\\Background\\shell\\AnyCode'))
# input()



from code_preset import ExistingCode

class CodeBuilder:

    def __init__(self, body_commands, script_dirs, funcs, imports, type):
        self.body_commands = body_commands
        self.script_dirs = list(set(script_dirs))
        self.funcs = funcs
        self.imports = list(set(imports))
        self.type = type.upper()

    def build_script_dirs(self):
        compiled_dirs = [f'sys.path.append("{x}")' for x in self.script_dirs]
        return '\n'.join(compiled_dirs)

    def build_imports(self):
        compiled_imports = [f'import {x}' for x in self.imports]
        return '\n'.join(compiled_imports)


    def compile(self):
        code_head = ExistingCode.CODE_HEAD.value
        script_dirs_code = self.build_script_dirs()
        imports_code = self.build_imports()
        class_dec = ExistingCode.CLASS_TEMPLATE.value
        class_funcs = '\n\n'.join(self.funcs)
        class_type = ExistingCode.FILE_ITEMS.value
        if self.type in ['DIRECTORY_BACKGROUND', 'DESKTOP_BACKGROUND']:
            class_type = ExistingCode.BACKGROUND_ITEMS.value
        class_body = '\n'.join(map(lambda x: '\t\t' + x, self.body_commands))


        code_skeleton = '''
{}
{}
{}
{}
{}
{}
{}
    '''.format(code_head, script_dirs_code, imports_code, class_dec, class_funcs, class_type, class_body)

        return code_skeleton



from enum import Enum


class ExistingCode(Enum):
    CODE_HEAD = '''
import gi
import sys
import os

gi.require_version('Nautilus', '3.0')

from gi.repository import Nautilus, GObject

# -------------------------------------------- #

try:
\tfrom urllib import unquote
except ImportError:
\tfrom urllib.parse import unquote
    '''

    CLASS_TEMPLATE = '''
class ExampleMenuProvider(GObject.GObject, Nautilus.MenuProvider):
\tdef __init__(self):
\t\tpass
'''

    METHOD_HANDLER_TEMPLATE = '''
\tdef {}(self, menu, files):
\t\tfilenames = [unquote(subFile.get_uri()[7:]) for subFile in files]
\t\t{}.{}({})

'''

    COMMAND_HANDLER_TEMPLATE = '''
\tdef {}(self, menu, files):
\t\tos.system('{}')

'''

    FILE_ITEMS = '\tdef get_file_items(self, window, files):'
    BACKGROUND_ITEMS = '\tdef get_background_items(self, window, files):'
    SUB_MENU = 'submenu{} = Nautilus.Menu()'
    MENU_ITEM = 'menuitem{} = Nautilus.MenuItem(name = "ExampleMenuProvider::{}", label="{}", tip = "{}", icon = "{}")'



import sys
import os

CONTEXT_SHORTCUTS = {
    'FILES': '*\\shell',
    'DIRECTORY': 'Directory\\shell',
    'DIRECTORY_BACKGROUND': 'Directory\\Background\\shell',
    'DESKTOP_BACKGROUND': 'DesktopBackground\\shell',
    'DRIVE': 'Drive\\shell'
}

COMMAND_PRESETS = {
    'python': sys.executable,
    'pythonw': os.path.join(os.path.dirname(sys.executable), 'pythonw.exe')
}


def context_registry_format(item):
    item = item.upper()
    if '.' in item:
        return item.lower()
    return CONTEXT_SHORTCUTS[item]


def command_preset_format(item):
    return COMMAND_PRESETS[item.lower()]

def create_file_select_command(func_name, func_file_name, func_dir_path):
    python_loc = sys.executable
    sys_section = f'''import sys; sys.path.insert(0, '{func_dir_path[2:]}')'''.replace('\\', '/')
    file_section = f'import {func_file_name}'
    dir_path = """' '.join(sys.argv[1:]) """
    func_section = f'{func_file_name}.{func_name}([{dir_path}])'
    python_portion = f'''"{python_loc}" -c "{sys_section}; {file_section}; {func_section}"'''
    full_command = '''{} \"%1\""'''.format(python_portion)
    
    return full_command


def create_directory_background_command(func_name, func_file_name, func_dir_path):
    python_loc = sys.executable
    sys_section = f'''import sys; import os; sys.path.insert(0, '{func_dir_path[2:]}')'''.replace('\\', '/')
    file_section = f'import {func_file_name}'
    dir_path = 'os.getcwd()'
    func_section = f'{func_file_name}.{func_name}([{dir_path}])'
    full_command = f'''"{python_loc}" -c "{sys_section}; {file_section}; {func_section}"'''

    return full_command

# "C:\\Users\\drale\\AppData\\Local\\Programs\\Python\\Python37-32\\python.exe" -c "import sys; sys.path.insert(0, '/Users/drale/Documents/Programming/dumb programs'); import test; test.main()" \"%1\""
# "C:\\Users\\drale\\AppData\\Local\\Programs\\Python\\Python37-32\\python.exe" -c "import sys; sys.path.insert(0, '/Users/drale/Documents/Programming/dumb programs'); import test; test.main()" \"%1\""



# Menus.py contextcommand
# from menus import *
#
# cc = ContextCommand('Directory\\Background\\shell\\test1', 'echo HELLO')
# cc.add_base_values({'Icon' : 'C:\\Users\\drale\\Pictures\\cmd.ico'})
# cc.compile()

# cc = ContextCommand('Test1', 'menus.py', preset='python')
# print(cc.command)

# ------------------------------------------------------------------


# Menus.py contextmenu
# from menus import *
#
# menu = ContextMenu('DIRECTORY_BACKGROUND', 'Test Menu 1')
# submenu1 = SubMenu('Sub Menu 1')
# submenu1.add_items([
#     ContextCommand('Cool Command 1', 'echo hello'),
#     ContextCommand('Cool Command 2', 'echo hello2'),
#     ContextCommand('Cool Command 3', 'echo hello3')
# ])
#
# submenu2 = SubMenu('Sub Menu 2')
# submenu2.add_items([
#     submenu1,
#     ContextCommand('Cool Command 4', 'echo hello2'),
#     ContextCommand('Cool Command 5', 'echo hello3')
# ])
#
# menu.add_items([submenu2, submenu1])
# menu.compile()

# ------------------------------------------------------------------

# Menus.py contextmenu order
# from menus import *
#
# menu = ContextMenu('DIRECTORY_BACKGROUND', 'Test Menu 1')
#
# for i in range(5):
#     submenu1 = SubMenu('Sub Menu {}'.format(i))
#     submenu1.add_items([
#         ContextCommand('Cool Command 1', 'echo hello'),
#     ])
#
#     if i != 4:
#         submenu1.add_items([
#             ContextCommand('Cool Command 3', 'echo hello'),
#             ContextCommand('Cool Command 2', 'echo hello2'),
#         ])
#
#     menu.add_items([submenu1])
#
# menu.compile()
# ------------------------------------------------------------------

# def lol():
#     print('hello')

# ------------------------------------------------------------------
# # registry_shortcuts.py
#
# from registry_shortcuts import *
#

# ------------------------------------------------------------------



from code_preset import ExistingCode
from code_builder import CodeBuilder
import os

class Variable:

    def __init__(self, name, code):
        self.name = name
        self.code = code


class NautilusMenu:

# Constructor, automatically handeled by menus.py
    def __init__(self, name, sub_items, type):
        self.name = name
        self.sub_items = sub_items
        self.type = type
        self.counter = 0

        self.commands = []
        self.script_dirs = []
        self.funcs = []
        self.imports = []

# Methods to create action code
    def append_item(self, menu, item):
        return "{}.append_item({})".format(menu, item)

    def set_submenu(self, item, menu):
        return '{}.set_submenu({})'.format(item, menu)

    def connect(self, item, func):
        return '{}.connect("activate", {}, files)'.format(item, func)

# Methods to create variable declarations

    def generate_menu(self):
        base_menu = ExistingCode.SUB_MENU.value
        base_menu = base_menu.format(self.counter)
        self.counter += 1

        return Variable(base_menu.split(' = ')[0], base_menu)

    def generate_item(self, name):
        base_command = ExistingCode.MENU_ITEM.value
        formatted_item = base_command.format(self.counter, name, name, '', '')
        self.counter += 1

        return Variable(formatted_item.split(' = ')[0], formatted_item)

    def generate_python_func(self, class_origin, class_func):
        func_name = 'method_handler{}'.format(self.counter)
        created_func = ExistingCode.METHOD_HANDLER_TEMPLATE.value.format(func_name, class_origin, class_func, 'filenames')

        self.counter += 1

        return Variable(f'self.{func_name}', created_func)

    def generate_command_func(self, command):
        func_name = 'method_handler{}'.format(self.counter)
        created_func = ExistingCode.COMMAND_HANDLER_TEMPLATE.value.format(func_name, command)

        self.counter += 1

        return Variable(f'self.{func_name}', created_func)





# Other misc methods to help out
    def get_next_item(self):
        val = self.generate_item('')
        self.counter -= 1

        return val.name



# Building the script body
    def build_script_body(self, name, items):
        top_item = self.generate_item(name)
        top_menu = self.generate_menu()
        submenu_com = self.set_submenu(top_item.name, top_menu.name)
        self.commands.append(top_item.code)
        self.commands.append(top_menu.code)
        self.commands.append(submenu_com)

        for item in items:
            if item.isMenu:
                subsubmenu_con = self.append_item(top_menu.name, self.get_next_item())
                self.build_script_body(item.name, item.sub_items)
                self.commands.append(subsubmenu_con)
            else:
                formatted_command = self.generate_item(item.name)
                self.commands.append(formatted_command.code)

                connected_func = None

                if item.python != None:
                    item_info = item.get_method_info()
                    connected_func = self.generate_python_func(item_info[1], item_info[0])
                    self.script_dirs.append(item_info[2])
                    self.imports.append(item_info[1])
                else:
                    connected_func =self.generate_command_func(item.command)
                    # connected_func = self.generate_func('os', 'system')

                self.funcs.append(connected_func.code)

                connected_command = self.connect(formatted_command.name, connected_func.name)
                self.commands.append(connected_command)

                formatted_command = self.append_item(top_menu.name, formatted_command.name)
                self.commands.append(formatted_command)


    def build_script(self):
        self.build_script_body(self.name, self.sub_items)
        self.commands.append('return menuitem0,')
        full_code = CodeBuilder(self.commands, self.script_dirs, self.funcs, self.imports, self.type).compile()

        return full_code

    def create_path(self, path, dir):
        new_dir = os.path.join(path, dir)
        if not os.path.exists(new_dir):
            os.makedirs(new_dir)
        return new_dir


    def compile(self):
        code = self.build_script()
        save_loc = os.path.join(os.path.expanduser('~'), '.local/share/')
        print(save_loc)
        save_loc = self.create_path(save_loc, 'nautilus-python')
        save_loc = self.create_path(save_loc, 'extensions')
        save_loc = os.path.join(save_loc, f'{self.name}.py')
        py_file = open(save_loc, 'w')
        py_file.write(code)
        py_file.close()




# top_command/
#     menu0/
#         menu1/
#               command1
#         command2
#     command3
#



import sys
import os
import platform

file_dir = os.path.dirname(__file__)
sys.path.append(file_dir)
sys.path.append(os.path.join(file_dir, 'windows'))
sys.path.append(os.path.join(file_dir, 'linux'))

if platform.system() == 'Linux':
    from linux import linux_menus
if platform.system() == 'Windows':
    from windows import windows_menus


import inspect


class ContextMenu:

    def __init__(self, name, type=None):
        self.name = name
        self.sub_items = []
        self.type = type
        self.isMenu = True # Needed to avoid circular imports



    def add_items(self, items):
        for item in items:
            self.sub_items.append(item)

    def compile(self):
        if platform.system() == 'Linux':
            linux_menus.NautilusMenu(self.name, self.sub_items, self.type).compile()
        if platform.system() == 'Windows':
            windows_menus.RegistryMenu(self.name, self.sub_items, self.type).compile()




class ContextCommand:

    def __init__(self, name, command=None, python=None):
        self.name = name
        self.command = command
        self.isMenu = False
        self.python = python

        if command != None and python != None:
            raise ValueError('both command and python cannot be defined')

    def get_platform_command(self):
        return self.command[platform.system().lower()]

    def get_method_info(self):
        func_file_path = os.path.abspath(inspect.getfile(self.python))

        func_dir_path = os.path.dirname(func_file_path)
        func_name = self.python.__name__
        func_file_name = os.path.splitext(os.path.basename(func_file_path))[0]

        return (func_name, func_file_name, func_dir_path)


class FastCommand:

    def __init__(self, name, type, command=None, python=None):
        self.name = name
        self.type = type
        self.command = command
        self.python = python

        if command != None and python != None:
            raise ValueError('both command and python cannot be defined')

    def get_method_info(self):
        func_file_path = os.path.abspath(inspect.getfile(self.python))

        func_dir_path = os.path.dirname(func_file_path)
        func_name = self.python.__name__
        func_file_name = os.path.splitext(os.path.basename(func_file_path))[0]

        return (func_name, func_file_name, func_dir_path)

    def compile(self):
        if platform.system() == 'Linux':
            linux_menus.NautilusMenu(self.name, [ContextCommand(self.name, command=self.command, python=self.python)], self.type).compile()
        if platform.system() == 'Windows':
            windows_menus.FastRegistryCommand(self.name, self.type, self.command, self.python).compile()




# import test1
# import test2
#
# fastCom = FastCommand('FooTest', 'FILES', python=test1.foo1)
# fastCom.compile()
#----------------------------------------------

# cm = ContextMenu('Foo', type='DIRECTORY_BACKGROUND')
#
# cm2 = ContextMenu('Foo2')
# cm3 = ContextMenu('Foo3')
# cm3.add_items([
#     ContextCommand('foo1', python=test1.foo1)
# ])
# cm2.add_items([
#     cm3,
#     ContextCommand('Command2', python=test1.foo2)
# ])
# cm.add_items([
#     cm2,
#     ContextCommand('Command3', python=test2.foo3)
# ])
#----------------------------------------------

# cm = ContextMenu('Foo')
# cm3 = ContextMenu('Foo3')
# cm3.add_items([
#     ContextCommand('Command1', 'ex')
# ])
#
# cm.add_items([
#     cm3,
#     ContextCommand('Command3', 'ex')
# ])
#
# print(cm.compile())

# top_command/
#     menu0/
#         menu1/
#               command1
#         command2
#     command3
#


# from menus import *
#
# menu = ContextMenu('DIRECTORY_BACKGROUND', 'Test Menu 1')
# submenu1 = SubMenu('Sub Menu 1')
# submenu1.add_items([
#     ContextCommand('Cool Command 1', 'echo hello'),
#     ContextCommand('Cool Command 2', 'echo hello2'),
#     ContextCommand('Cool Command 3', 'echo hello3')
# ])
#
# submenu2 = SubMenu('Sub Menu 2')
# submenu2.add_items([
#     submenu1,
#     ContextCommand('Cool Command 4', 'echo hello2'),
#     ContextCommand('Cool Command 5', 'echo hello3')
# ])
#
# menu.add_items([submenu2, submenu1])
# menu.compile()



import registry_shortcuts as reg
import advanced_reg_config as arc
import os



# ------------------------------------------------------------------


class RegistryMenu:

    def __init__(self, name, sub_items, type):
        self.name = name
        self.sub_items = sub_items
        self.type = type.upper()
        self.path = arc.context_registry_format(type)

    def create_menu(self, name, path):
        key_path = os.path.join(path, name)
        reg.create_key(key_path)

        reg.set_key_value(key_path, 'MUIVerb', name)
        reg.set_key_value(key_path, 'subcommands', '')

        key_shell_path = os.path.join(key_path, 'shell')
        reg.create_key(key_shell_path)

        return key_shell_path

    def create_command(self, name, path, command):
        key_path = os.path.join(path, name)
        reg.create_key(key_path)
        reg.set_key_value(key_path, '', name)

        command_path = os.path.join(key_path, 'command')
        reg.create_key(command_path)
        reg.set_key_value(command_path, '', command)






    def compile(self, items=None, path=None):
        if items == None:
            reg.run_admin()
            items = self.sub_items
            path = self.create_menu(self.name, self.path)

        for item in items:
            if item.isMenu:
                submenu_path = self.create_menu(item.name, path)
                self.compile(items=item.sub_items,path=submenu_path)
            else:
                if item.command == None:
                    func_name, func_file_name, func_dir_path = item.get_method_info()
                    new_command = None
                    if self.type in ['DIRECTORY_BACKGROUND', 'DESKTOP_BACKGROUND']:
                        new_command = arc.create_directory_background_command(func_name, func_file_name, func_dir_path)
                    else:
                        new_command = arc.create_file_select_command(func_name, func_file_name, func_dir_path)
                    self.create_command(item.name, path, new_command)
                else:
                    self.create_command(item.name, path, item.command)



class FastRegistryCommand:

    def __init__(self, name, type, command, python):
        self.name = name
        self.type = type
        self.path = arc.context_registry_format(type)
        self.command = command
        self.python = python

    def get_method_info(self):
        import inspect

        func_file_path = os.path.abspath(inspect.getfile(self.python))

        func_dir_path = os.path.dirname(func_file_path)
        func_name = self.python.__name__
        func_file_name = os.path.splitext(os.path.basename(func_file_path))[0]

        return (func_name, func_file_name, func_dir_path)

    def compile(self):
        reg.run_admin()

        key_path = os.path.join(self.path, self.name)
        reg.create_key(key_path)

        command_path = os.path.join(key_path, 'command')
        reg.create_key(command_path)

        new_command = None


        if self.command == None:
            func_name, func_file_name, func_dir_path = self.get_method_info()
            if self.type in ['DIRECTORY_BACKGROUND', 'DESKTOP_BACKGROUND']:
                new_command = arc.create_directory_background_command(func_name, func_file_name, func_dir_path)
            else:
                new_command = arc.create_file_select_command(func_name, func_file_name, func_dir_path)
        else:
            new_command = self.command

        reg.set_key_value(command_path, '', new_command)
